using System;
using Overdash.Combat;
using Overdash.Global;
using UnityEngine;

namespace Overdash.Player
{
    [RequireComponent(typeof(PlayerMovement), typeof(PlayerCombat), typeof(PlayerSound))]
    [RequireComponent(typeof(PlayerPickupField), typeof(Rigidbody2D))]
    public class PlayerCharacter : MonoBehaviour
    {
        // Sub-components
        public PlayerMovement movement { get; private set; }
        public PlayerCombat combat { get; private set; }
        public PlayerSound sound { get; private set; }
        public PlayerPickupField pickupField { get; private set; }
        public Rigidbody2D rigidBody { get; private set; }
        
        // Controlled by:
        public Player player { get; private set; }
        
        // Properties
        [SerializeField] private GameObject aimIndicator;

        // Events
        public event Action<PlayerDeathInfo> OnKilled;

        private AmmoType currentDisplayedAmmoType;
        
        private void Awake()
        {
            // Get sub-components
            rigidBody = GetComponent<Rigidbody2D>();
            movement = GetComponent<PlayerMovement>();
            combat = GetComponent<PlayerCombat>();
            sound = GetComponent<PlayerSound>();
            pickupField = GetComponent<PlayerPickupField>();
            
            // Disabled movement / combat unless round is in progress
            if (RoundLoop.state != RoundLoop.State.InProgress)
            {
                movement.enabled = false;
                combat.enabled = false;
            }
            
            // Subscribe to events
            RoundLoop.OnRoundStart += EnableCharacter;
        }

        private void EnableCharacter()
        {
            combat.enabled = true;
            movement.enabled = true;
        }

        public void Initialize(Player player)
        {
            this.player = player;
            
            // Set player color
            GetComponentInChildren<SpriteRenderer>().material.color = player.color;
            GetComponentInChildren<SpriteRenderer>().material.color = player.color;
        }

        private void FixedUpdate()
        {
            updateAimIndicator();
        }

        private void updateAimIndicator()
        {
            Vector2 lookDirection = GetComponent<PlayerMovement>().lookDirection == new Vector2(0, 0) ? GetComponent<PlayerMovement>().direction * Vector2.right : GetComponent<PlayerMovement>().lookDirection;
            aimIndicator.GetComponentInChildren<SpriteRenderer>().material.color = player.color;
            aimIndicator.transform.position = transform.position + (Vector3)lookDirection * 1.6f;
            aimIndicator.transform.rotation = Quaternion.LookRotation(Vector3.forward, lookDirection);
        }

        public void Kill(Player byPlayer)
        {
            var deathInfo = new PlayerDeathInfo
            {
                KilledPlayer = player,
                KilledByPlayer = byPlayer
            };
            OnKilled?.Invoke(deathInfo);
            Destroy(gameObject);
        }

        private void OnDestroy() => RoundLoop.OnRoundStart -= EnableCharacter;
    }
}
