﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Overdash.Combat;
using Overdash.Combat.Projectiles;
using Overdash.General;
using Overdash.Global;
using Overdash.Levels;
using UnityEngine;
using UnityEngine.InputSystem;
using Random = UnityEngine.Random;

namespace Overdash.Player
{
    [RequireComponent(typeof(PlayerCharacter), typeof(PlayerMovement))]
    public class PlayerCombat : MonoBehaviour
    {
        [Header("Settings")]
        [SerializeField] private float fireDistance;
        [SerializeField] private int maxAmmo;
        
        // Events
        public event Action<AmmoType> OnAmmunitionAdded;
        public event Action<AmmoType> OnShoot;
        public event Action OnShootEmptyClip;

        // References
        private PlayerCharacter playerCharacter;
        private PlayerInput playerInput;
        private PlayerMovement playerMovement;

        // Properties
        public readonly Stack<AmmoType> Ammunition = new();

        private void Awake()
        {
            playerCharacter = GetComponent<PlayerCharacter>();
            playerMovement = GetComponent<PlayerMovement>();
            
            GetComponent<PlayerPickupField>().PickupCallbacks.Add(TryPickupAmmo);
            
            // Set ammunition
            var startingAmmo = LevelManager.currentLevel.customStartingAmmoEnabled 
                                                    ? LevelManager.currentLevel.customStartingAmmo 
                                                    : GameLoop.parameters.playerStartingAmmo;
            
            if ((Levels.Modifiers.OneBullet.isActive || Levels.Modifiers.InfiniteAmmo.isActive) && startingAmmo.Count > 0)
                Ammunition.Push(startingAmmo.First());
            else
                foreach (var bullet in startingAmmo)
                    Ammunition.Push(bullet);
        }

        private void Start()
        {
            playerInput = playerCharacter.player.input;
            playerInput.actions["Fire"].performed += Fire;
            playerCharacter.OnKilled += OnPlayerKilled;
        }   
        
        private bool TryPickupAmmo(PickupItem pickupItem)
        {
            // If pickup is ammo
            if (pickupItem.TryGetComponent(out AmmoPickup ammoPickup))
            {
                // Skip if player already has max ammo
                if (Ammunition.Count >= maxAmmo)
                    return false;
                    
                // Skip if OneBullet active and player already has ammo
                if (Levels.Modifiers.OneBullet.isActive && GetComponent<PlayerCombat>().Ammunition.Count > 0)
                    return false;
                    
                // Add ammo
                AddAmmunition(ammoPickup.ammoType);
                
                return true;
            }
            return false;
        }

        private void Fire(InputAction.CallbackContext _)
        {
            // If player has ammo
            if ((Levels.Modifiers.InfiniteAmmo.isActive && Ammunition.TryPeek(out var bullet)) || Ammunition.TryPop(out bullet))
            {
                var position = transform.position + (Vector3)playerMovement.lookDirection * fireDistance;
                var rotation = Quaternion.LookRotation(Vector3.forward, playerMovement.lookDirection == Vector2.zero ? new Vector2(playerMovement.direction, 0) : playerMovement.lookDirection);
                
                var projectile = Instantiate(bullet.projectilePrefab, position, rotation).GetComponent<ProjectileBase>();
                StartCoroutine(RestoreCollision(projectile.GetComponent<Collider2D>(), playerCharacter.GetComponent<Collider2D>()));
                projectile.OnHitPlayer += OnFiredProjectileHitPlayer;
                projectile.Shoot(playerCharacter.player);
                OnShoot?.Invoke(bullet);
            }
            else
            {
                OnShootEmptyClip?.Invoke();
            }
        }

        private void OnDestroy()
        {
            if (playerInput == null) return;
            playerInput.actions["Fire"].performed -= Fire;
            playerCharacter.OnKilled -= OnPlayerKilled;
        }

        private static void OnFiredProjectileHitPlayer(ProjectilePlayerHit playerHit)
        {
            // Hit another player
            if (playerHit.Receiver != playerHit.Shooter)
                Scoreboard.AddPoint(playerHit.Shooter);
            
            // Hit self
            else
                Scoreboard.RemovePoint(playerHit.Shooter);
        }
        
        private void OnPlayerKilled(PlayerDeathInfo _)
        {
            DropAllAmmo();
        }

        private void DropAllAmmo()
        {
            while (Ammunition.Count > 0)
            {
                var bullet = Ammunition.Pop();
                var position = (Vector2)transform.position + Random.insideUnitCircle;
                var projectile = Instantiate(bullet.projectilePrefab, position, Quaternion.AngleAxis(Random.value * 180, Vector3.forward)).GetComponent<ProjectileBase>();
                var projectileRigidbody = projectile.GetComponent<Rigidbody2D>();
                projectileRigidbody.AddForce(playerCharacter.rigidBody.velocity + Random.insideUnitCircle * 5, ForceMode2D.Impulse);
                projectileRigidbody.AddTorque(Random.Range(-5, 5), ForceMode2D.Impulse);
                projectileRigidbody.constraints = RigidbodyConstraints2D.None;
                PropsManager.PropsList.Add(projectile.gameObject);
            }
        }
        
        public void AddAmmunition(AmmoType ammoType)
        {
            if (Levels.Modifiers.OneBullet.isActive && Ammunition.Count > 0) return;
            Ammunition.Push(ammoType);
            OnAmmunitionAdded?.Invoke(ammoType);
        }

        IEnumerator RestoreCollision(Collider2D projectileCollider, Collider2D playerCollider)
        {
            Physics2D.IgnoreCollision(projectileCollider, playerCollider);
            yield return new WaitForSeconds(0.2f);
            Physics2D.IgnoreCollision(projectileCollider, playerCollider, false);
        }

    }
}