using System;
using System.Collections.Generic;
using System.Linq;
using Overdash.Common;
using Overdash.Global;
using Overdash.Levels;
using UnityEngine;
using UnityEngine.InputSystem;

namespace Overdash.Player
{
    [RequireComponent(typeof(PlayerInputManager))]
    public class PlayerManager : MonoBehaviourSingleton<PlayerManager>
    {
        [SerializeField] private GameObject playerCharacterPrefab;
        [SerializeField] private List<Color> playerColors = new();
        
        
        public static event Action<Player> OnPlayerJoined;
        public static event Action<PlayerDeathInfo> OnPlayerKilled; 
        public static List<Player> players { get; private set; } = new();
        public static List<Player> alivePlayers { get; private set; } = new();
        
        
        private static GameObject _playerCharacterPrefab;
         
        protected override void Awake()
        {
            base.Awake();
            _playerCharacterPrefab = playerCharacterPrefab;
            
            var playerInputManager = GetComponent<PlayerInputManager>();
            playerInputManager.onPlayerJoined += OnPlayerInputRegistered;
            playerInputManager.onPlayerLeft += OnPlayerInputUnregister;
            
            LevelManager.OnLevelLoaded += _ =>
            {
                foreach (var player in players)
                    SpawnPlayerCharacter(player);
            };
            
            RoundLoop.OnRoundEnd += () =>
            {
                foreach (var player in alivePlayers.ToList())
                {
                    Destroy(player.controlledCharacter.gameObject);
                    alivePlayers.Remove(player);
                }
            };
            
            // This runs Scoreboard's static constructor before first player joins
            // https://learn.microsoft.com/en-us/dotnet/csharp/programming-guide/classes-and-structs/static-constructors
            Scoreboard.Score.Clear(); 
        }

        private void OnPlayerInputRegistered(PlayerInput playerInput)
        {
            Debug.Log($"Player {playerInput.playerIndex} joined");
            var player = playerInput.GetComponent<Player>();
            players.Add(player);
            player.Initialize(players.Count, playerColors[players.Count - 1]);
            
            // TODO: Remove, new player should not be able to join mid-game
            SpawnPlayerCharacter(player);
            
            OnPlayerJoined?.Invoke(player);
        }
        
        private void OnPlayerInputUnregister(PlayerInput playerInput)
        {
            var player = playerInput.GetComponent<Player>();
            players.Remove(player);
            Destroy(player.gameObject);
        }
        
        public static PlayerCharacter SpawnPlayerCharacter(Player player)
        {
            var spawnPoint = FindObjectOfType<Level>().spawnPoints.First(x => !x.isOccupied);
            spawnPoint.isOccupied = true;
            var playerCharacter = Instantiate(_playerCharacterPrefab, spawnPoint.transform.position, Quaternion.identity).GetComponent<PlayerCharacter>();
            playerCharacter.OnKilled += OnPlayerKilledCallback;
            playerCharacter.Initialize(player);
            player.AssignCharacter(playerCharacter);
            alivePlayers.Add(player);
            return playerCharacter;
        }

        private static void OnPlayerKilledCallback(PlayerDeathInfo deathInfo)
        {
            alivePlayers.Remove(deathInfo.KilledPlayer);
            OnPlayerKilled?.Invoke(deathInfo);
        }
    }
}