using System;
using System.Collections;
using UnityEngine;
using UnityEngine.InputSystem;

namespace Overdash.Player
{
    [RequireComponent(typeof(Rigidbody2D))]
    public class PlayerMovement : MonoBehaviour
    {
        // References
        private PlayerInput playerInput;
        private InputAction moveAction;
        private new Rigidbody2D rigidbody;

        // Properties
        [field: SerializeField] public int direction { get; private set; }
        [field: SerializeField] public Vector2 lookDirection { get; private set; }

        [Header("Movement Settings")]
        [Range(0.001f, 1)]
        [SerializeField] private float acceleration;
        [SerializeField] private float maxVelocity;
        [SerializeField] private float AbsoluteMaxVerticalVelocity; //for infinity falling
        [SerializeField] private float jumpForce;
        [SerializeField] private int airJumpChargesMaximum;
        //[SerializeField] private float dashCooldown; // neviem ci chceme skorej cooldown alebo charge
        [SerializeField] private float dashDuration;
        [SerializeField] private float dashSpeed;
        [SerializeField] private int dashChargesMaximum;
        [SerializeField] private float duckCollider_Vs_NormalCollider_HeightRatio; //0..1
        [Header("Ground Detection Settings")]
        [SerializeField] private Vector2 groundedCheckerSize;
        [SerializeField] private float groundCheckerCastDistance;
        [SerializeField] private LayerMask groundLayer;
        [SerializeField] private Vector2 slideCheckerSize;
        [SerializeField] private float slideCheckerCastDistance;
        [SerializeField] private float slideCheckerVerticalOffset;
        [SerializeField] private float slideForce;//force put into the wall when sliding 0...1
        [Header("Animator")]
        [SerializeField] private Animator animator;

        private int jumpsRemaining;
        private bool canRefreshJumps = true; //immediately after jumping, it still detects as being grounded and would refresh jumpsRemaining and/or trigger additional jumps
        private bool canRefreshDashes = true;
        private bool canDash = true; // cooldown for dash to prevent spamming on ground/walls, but still recharge it via CanRefreshDashes
        private bool sliding = false;
        private bool grounded = false;
        private int dashesRemaining;
        private bool ducking = false;
        private Coroutine endDashCoroutine;
        private Vector2 normalColliderSize;
        private Vector2 dashingDirection;
        public  bool dashing { get; private set; } = false;

        private SpriteRenderer spriteRenderer = null;

        // Events for audio
        public event Action OnDash;
        public event Action OnJumpGround;
        public event Action OnJumpAir;
        public event Action OnHitGround;
        public event Action OnSlide;
        public event Action OnEndSlide;



        private void Awake()
        {
            rigidbody = GetComponent<Rigidbody2D>();
        }

        private void Start()
        {
            playerInput = GetComponent<PlayerCharacter>().player.input;
            moveAction = playerInput.actions["Move"];
            playerInput.actions["Dash"].performed += Dash;
            playerInput.actions["Jump"].performed += Jump;
            jumpsRemaining = airJumpChargesMaximum;
            dashesRemaining = dashChargesMaximum;
            normalColliderSize = GetComponent<BoxCollider2D>().size;

            spriteRenderer = GetComponentInChildren<SpriteRenderer>();

            GetComponent<Rigidbody2D>().collisionDetectionMode = CollisionDetectionMode2D.Continuous; // discrete je default a je horsie ako toto
        }

        private void FixedUpdate()//bit messy
        {
            Vector2 moveVec = moveAction.ReadValue<Vector2>();
            bool wasGrounded = grounded;
            grounded = isGrounded();
            if (grounded && !wasGrounded)
            {
                OnHitGround?.Invoke();
            }
            if (grounded)
            {
                if (sliding) OnEndSlide?.Invoke();
                sliding = false;
                if (canRefreshJumps)  jumpsRemaining = airJumpChargesMaximum;
                if (canRefreshDashes) dashesRemaining = dashChargesMaximum;

                if (moveVec.y < -0.5f)
                {
                    Duck();
                }
                else
                {
                    endDuck();
                }
            }
            else
            {   bool wasSliding = sliding;
                sliding = isSliding();
                if (sliding && !wasSliding) OnSlide?.Invoke();
                if (!sliding && wasSliding) OnEndSlide?.Invoke();
                endDuck();
            }
            if (!grounded && sliding && Math.Sign(moveVec.x) == direction)
            {
                moveVec.x = Math.Sign(moveVec.x) * slideForce;
                jumpsRemaining = airJumpChargesMaximum;
                dashesRemaining = dashChargesMaximum;
            }

            UpdateAnimator(moveVec);

            if (spriteRenderer != null && Mathf.Abs(moveVec.x) > 0.0f)
            {
                spriteRenderer.flipX = moveVec.x > 0.0f;
            }

            if (dashing)
            {
                DashMove();
            }
            else
            {
                Move(moveVec);
            }
        }

        private void OnDestroy()
        {
            if (playerInput == null) return;
            playerInput.actions["Jump"].performed -= Jump;
            playerInput.actions["Dash"].performed -= Dash;
        }

        private void Move(Vector2 input)
        {
            lookDirection = input.normalized;
            this.direction = input.x > 0 ? 1 : input.x < 0 ? -1 : direction;
            rigidbody.velocity = new Vector2(Mathf.Lerp(rigidbody.velocity.x, input.x * maxVelocity, acceleration), rigidbody.velocity.y);
            if (Math.Abs(rigidbody.velocity.y) > AbsoluteMaxVerticalVelocity)
            {
                rigidbody.velocity = new Vector2(rigidbody.velocity.x, Math.Sign(rigidbody.velocity.y) * AbsoluteMaxVerticalVelocity);
            }

        }
        private void DashMove()
        {
            rigidbody.velocity = dashingDirection * dashSpeed;
        }

        private bool isGrounded()
        {
            return Physics2D.BoxCast(transform.position, groundedCheckerSize, 0, -Vector2.up, groundCheckerCastDistance, groundLayer);
        }

        private bool isSliding()
        {
            return Physics2D.BoxCast(transform.position + transform.up * slideCheckerVerticalOffset, slideCheckerSize, 0, Vector2.right, slideCheckerCastDistance, groundLayer)
                || Physics2D.BoxCast(transform.position + transform.up * slideCheckerVerticalOffset, slideCheckerSize, 0, -Vector2.right, slideCheckerCastDistance, groundLayer);
        }

        private void OnDrawGizmos()//check player prefab, pocas hry nefunguje..
        {
            Gizmos.color = Color.red;
            Gizmos.DrawWireCube(transform.position - transform.up * groundCheckerCastDistance, groundedCheckerSize);
            Gizmos.DrawWireCube(transform.position + transform.right * slideCheckerCastDistance + transform.up * slideCheckerVerticalOffset, slideCheckerSize);
            Gizmos.DrawWireCube(transform.position - transform.right * slideCheckerCastDistance + transform.up * slideCheckerVerticalOffset, slideCheckerSize);
        }


        private void Jump(InputAction.CallbackContext _)
        {
            forceEndDash();
            endDuck();
            if (isGrounded() || sliding)
            {
                rigidbody.velocity = new Vector2(Math.Abs(rigidbody.velocity.x) > maxVelocity ? Mathf.Sign(rigidbody.velocity.x) * maxVelocity : rigidbody.velocity.x, jumpForce);
                if (grounded) OnJumpGround?.Invoke();
                else OnJumpAir?.Invoke();
            }
            else if(jumpsRemaining > 0 || Levels.Modifiers.InfiniteJumps.isActive)
            { 
                jumpsRemaining -= 1;
                rigidbody.velocity = new Vector2(Math.Abs(rigidbody.velocity.x) > maxVelocity ? Mathf.Sign(rigidbody.velocity.x) * maxVelocity : rigidbody.velocity.x, jumpForce);

                
                StartCoroutine(CycleCanRefreshJumps());
                if (grounded) OnJumpGround?.Invoke();
                else OnJumpAir?.Invoke();
                animator.SetTrigger("jump");
            }
        }

        private void Dash(InputAction.CallbackContext _)
        {
            if (dashesRemaining == 0 || dashing || !canDash || Levels.Modifiers.NoDashing.isActive) return;
            endDashCoroutine = StartCoroutine(EndDash());
            dashingDirection = lookDirection.normalized;
            if (dashingDirection == Vector2.zero || (dashingDirection.y < 0 && isGrounded())) dashingDirection = Vector2.right * direction;

            StartCoroutine(CycleCanRefreshDashes());
            StartCoroutine(CycleCanDash());
            dashing = true;
            dashesRemaining -= 1;
            OnDash?.Invoke();

            rigidbody.velocity = Vector2.right * dashSpeed * direction;
        }

        private void Duck()
        {
            if (!isGrounded() || dashing || ducking) return;
            ducking = true;
            //change collider to be shorter
            GetComponent<BoxCollider2D>().size = new Vector2(GetComponent<BoxCollider2D>().size.x, GetComponent<BoxCollider2D>().size.y * duckCollider_Vs_NormalCollider_HeightRatio);
            GetComponent<BoxCollider2D>().offset = new Vector2(0, -(normalColliderSize.y - normalColliderSize.y * duckCollider_Vs_NormalCollider_HeightRatio) / 2);
        }

        private void endDuck()
        {
            if (!ducking) return;
            ducking = false;
            //change collider to be normal
            GetComponent<BoxCollider2D>().size = normalColliderSize;
            GetComponent<BoxCollider2D>().offset = new Vector2(0, 0);
        }
        private void UpdateAnimator(Vector2 moveVec)
        {
            animator.SetFloat("walking", Mathf.Abs(moveVec.x));
            animator.SetBool("ducking", ducking);
            animator.SetBool("grounded", isGrounded());
            animator.SetBool("sliding", sliding);
            animator.SetBool("dashing", dashing);
        }

        //-------coroutines related stuff------------

        private void forceEndDash()
        {
            if (!dashing) return;
            if (endDashCoroutine != null)
            {
                StopCoroutine(endDashCoroutine);
                endDashCoroutine = null;
            }
            dashing = false;
            SetSpeedsAfterDash();
        }
        IEnumerator EndDash()
        {
            yield return new WaitForSeconds(dashDuration);
            dashing = false;
            SetSpeedsAfterDash();

        }
        private void SetSpeedsAfterDash()
        {
            if (Math.Abs(rigidbody.velocity.x) > maxVelocity)
            {
                rigidbody.velocity = new Vector2(maxVelocity * direction, rigidbody.velocity.y);
            }
            if (Math.Abs(rigidbody.velocity.y) > maxVelocity / 2)//arbitrary number 2 and 4
            {
                rigidbody.velocity = new Vector2(rigidbody.velocity.x, Math.Sign(rigidbody.velocity.y) * maxVelocity / 4);
            }
        }
        IEnumerator CycleCanRefreshJumps()
        {
            canRefreshJumps = false;
            yield return new WaitForSeconds(0.1f);
            canRefreshJumps = true;
        }
        IEnumerator CycleCanRefreshDashes()
        {
            canRefreshDashes = false;
            yield return new WaitForSeconds(0.1f);
            canRefreshDashes = true;
        }
        IEnumerator CycleCanDash()
        {
            canDash = false;
            yield return new WaitForSeconds(0.3f);
            canDash = true;
        }
    }
}